@@ -143,5 +143,7 @@ extern char *no_nextprop(int node, char *name);
 /* PROM Edit Device Structures (prior to booting Linux) */
 extern void prop_add(struct node *n, struct prop *newprop);
 extern void prop_del(struct node *n, char *name);
+extern void prop_add_dup(struct node *n, struct prop_std *buf,
+			  int size, struct prop_std **lastprop);
 
 #endif
