@@ -27,6 +27,9 @@
 #include <gptimer.h>
 #include <irqmp.h>
 #include <linkscript.h>
+#ifdef CONFIG_WATCHDOG
+#include <prom_watchdog.h>
+#endif
 
 static void leon_reboot(char *bcommand);
 static void leon_halt(void);
@@ -324,8 +327,13 @@ char *prom_strdup(char *str)
 
 static void leon_reboot(char *bcommand)
 {
+#ifdef CONFIG_WATCHDOG
+	outstr("PROM: Reboot\n");
+	watchdog_set_timeout(10);
+#else
 	outstr("PROM: Reboot  (not implemented)\n");
 
+#endif
 	while (1)
 		;
 }
@@ -395,6 +403,11 @@ static int leon_nbputchar(int c)
 
 	BYPASS_STORE_PA(&b->data, c & 0xff);
 	BYPASS_STORE_PA(&b->ctrl, ctrl);
+
+#ifdef CONFIG_WATCHDOG
+	watchdog_ping(PING_ID_UART);
+#endif
+
 	return 1; /* One char transmitted */
 }
 #endif
