@@ -15,6 +15,12 @@
 #define GR740_REV0_SYSTEMID 0x07401038
 #define GR740_REV1_SYSTEMID 0x07401039
 
+static inline int amba_systemid_is_gr740(void)
+{
+	return (amba_systemid == GR740_REV0_SYSTEMID) ||
+	       (amba_systemid == GR740_REV1_SYSTEMID);
+}
+
 #ifdef CONFIG_AMPOPTS
 struct amp_opt {
 	short idx, val;
@@ -103,6 +109,18 @@ int irqmp_dev(ambapp_dev *dev)
 	return 0;
 }
 
+void node_add_child(struct node *node, struct node *child);
+int node_has_child(struct node *node, const char *child_name, int name_len, struct node **child);
+
+void ambapp_prop_init(struct prop_std *buf, const char *prop_str);
+int ambapp_prop_int(struct prop_std *buf, const char *prop_str, int value);
+int ambapp_prop_str(struct prop_std *buf, const char *prop_str, const char *val_str);
+int ambapp_prop_uintarray(struct prop_std *buf, const char *prop_str,
+			  unsigned int *array, int len);
+int ambapp_prop_str_array(struct prop_std *buf, const char *prop_str,
+			  const char *val_array[]);
+int ambapp_prop_empty(struct prop_std *buf, const char *prop_str);
+
 static int gptimer_inited = 0;
 
 /* Fixup IRQ of GPTIMER APBSLV interface */
@@ -128,6 +146,101 @@ int gptimer_apb_fixup(ambapp_dev *dev)
 	return 0;
 }
 
+#ifdef CONFIG_WATCHDOG
+
+#define GPTIMER_WATCHDOG_GR740_HW_TIMEOUT 2000
+
+void gptimer_add_watchdog_node(struct node *node)
+{
+	struct node *n;
+	int size;
+	unsigned hw_timeout_ms = 0;
+	unsigned int prop_buf[64];
+	struct prop_std *buf = (struct prop_std *)prop_buf;
+
+#ifdef HASXML
+	/* Check if there already exist a watchdog node. In that case skip
+	 * adding the node as it most likely has been added from xml
+	 */
+	if (node_has_child(node, prop_str_watchdog, 16, NULL))
+		return;
+#endif
+
+	/* Set the hw timeout. */
+	if (amba_systemid_is_gr740()) {
+		/* GR740 has a PLL watchdog with a timeout of 2 seconds */
+		hw_timeout_ms = GPTIMER_WATCHDOG_GR740_HW_TIMEOUT;
+	} else {
+		uint32_t config;
+		int ntimers;
+		/* Get the pre-configured reload value
+		 * from the watchdog timer which is assumed to be
+		 * configured on boot.
+		 */
+		config = BYPASS_LOAD_PA(&gptimer->config);
+		ntimers = config & 0x7;
+		hw_timeout_ms =
+			BYPASS_LOAD_PA(&gptimer->timer[ntimers - 1].rld);
+		hw_timeout_ms /= 1000;
+	}
+	/* The hw_timeout_ms needs to be set, otherwise skip. */
+	if (!hw_timeout_ms)
+		return;
+
+	/* Create a watchdog node and add it:
+	 * watchdog {
+	 *    compatible = "gaisler,gptimer-wdt";
+	 *    gaisler,max-hw-timeout-ms = 2000;
+	 *    timeout-sec = 10;
+	 *    name = "watchdog";
+	 * };
+	 */
+
+	n = (struct node *)prom_malloc(sizeof(struct node));
+	n->child = NULL;
+	n->sibling = NULL;
+	n->props = NULL;
+
+	/* Name of the node */
+	size = ambapp_prop_str(buf, prop_str_name, prop_str_watchdog);
+	prop_add_dup(n, buf, size, NULL);
+
+	/* Set the compatible string so the driver can match */
+	size = ambapp_prop_str(buf, prop_str_compatible, "gaisler,gptimer-wdt");
+	prop_add_dup(n, buf, size, NULL);
+
+#ifdef CONFIG_WATCHDOG_USER_TIMEOUT
+	/* Override the user timeout if it is specified */
+	size = ambapp_prop_int(buf, prop_str_timeout_sec,
+			CONFIG_WATCHDOG_USER_TIMEOUT);
+	prop_add_dup(n, buf, size, NULL);
+#endif
+
+	/* Set the hw timeout of the watchdog */
+	size = ambapp_prop_int(buf, prop_str_hw_timeout_ms, hw_timeout_ms);
+	buf->options |= PO_END;
+	prop_add_dup(n, buf, size, NULL);
+
+	/* Finally, add the watchdog node to the parent node */
+	node_add_child(node, n);
+}
+
+int gptimer_fixup_node(ambapp_core *core, struct node *node)
+{
+	/* Use static as core subindex is not generally available. */
+	static int numnode = 0;
+
+	if (numnode == 0)
+		gptimer_add_watchdog_node(node);
+
+	numnode++;
+
+	return 0;
+}
+#else
+#define gptimer_fixup_node NULL
+#endif
+
 struct mutable_ambapp_pnp_ahb {
 	unsigned int	id;		/* VENDOR, DEVICE, VER, IRQ, */
 	unsigned int	custom[3];
@@ -171,15 +284,6 @@ int canoc_ahb_fixup_virtual(ambapp_dev *dev, void *axb, void *bus, void **pdevp_
 	return 1;
 }
 
-void ambapp_prop_init(struct prop_std *buf, const char *prop_str);
-int ambapp_prop_int(struct prop_std *buf, const char *prop_str, int value);
-int ambapp_prop_str(struct prop_std *buf, const char *prop_str, const char *val_str);
-int ambapp_prop_uintarray(struct prop_std *buf, const char *prop_str,
-			  unsigned int *array, int len);
-int ambapp_prop_str_array(struct prop_std *buf, const char *prop_str,
-			  const char *val_array[]);
-int ambapp_prop_empty(struct prop_std *buf, const char *prop_str);
-
 const char *canoc_compatibles[] = {
 	"gaisler,can_oc",
 	"aeroflexgaisler,can_oc",
@@ -680,38 +784,47 @@ struct amba_driver {
 	int (*fixup_virtual_dev)(ambapp_dev *dev, void *axb, void *bus,
 				 void **pdevp);
 	ambapp_core *cores;
+	int (*fixup_node)(ambapp_core *core, struct node *node);
 };
 
 /* Driver for a special version */
-#define AMBAPP_DRV_VER(ven, dev, vmin, vmax, fixdev, fixcore, fixprop, init) \
-	{ven, dev, vmin, vmax, fixdev, fixcore, fixprop, init, NULL, NULL, NULL}
+#define AMBAPP_DRV_VER(ven, dev, vmin, vmax, fixdev, fixcore, fixprop, init, fixnode) \
+	{ven, dev, vmin, vmax, fixdev, fixcore, fixprop, init, NULL, NULL, NULL, fixnode}
 /* Driver for all versions */
-#define AMBAPP_DRV(ven, dev, fixdev, fixcore, fixprop, init) \
-	{ven, dev, 0, 0xff, fixdev, fixcore, fixprop, init, NULL, NULL, NULL}
+#define AMBAPP_DRV(ven, dev, fixdev, fixcore, fixprop, init, fixnode) \
+	{ven, dev, 0, 0xff, fixdev, fixcore, fixprop, init, NULL, NULL, NULL, fixnode}
 /* Driver for all versions with extra fixup functions */
 #define AMBAPP_DRV_EXTRA_FIXUP(ven, dev, fixdev, fixcore, fixprop, init, \
-			       fixdevcnt, fixvirtdev)                   \
+			       fixdevcnt, fixvirtdev, fixnode)                   \
 	{ven, dev, 0, 0xff, fixdev, fixcore, fixprop, init,             \
-			fixdevcnt, fixvirtdev, NULL}
+			fixdevcnt, fixvirtdev, NULL, fixnode}
 
 struct amba_driver amba_drivers[] = {
  /*					   dev, core, prop, init */
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_IRQMP, irqmp_dev, NULL, NULL, NULL),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPTIMER, gptimer_apb_fixup, NULL, NULL, gptimer_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON5, NULL, NULL, NULL, leon_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON4, NULL, NULL, NULL, leon_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3, NULL, NULL, NULL, leon_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3FT, NULL, NULL, NULL, leon_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_APBUART, apbuart_apb_fixup, NULL, NULL, NULL),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_ETHMAC, NULL, NULL, greth_fixup_prop, greth_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_I2CMST, NULL, NULL, i2c_fixup_prop, NULL),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_SPICTRL, NULL, NULL, spi_fixup_prop, NULL),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPIO, NULL, grgpio_fixup_core, grgpio_fixup_prop, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_IRQMP, irqmp_dev, NULL, NULL, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPTIMER, gptimer_apb_fixup, NULL,    \
+			NULL, gptimer_init, gptimer_fixup_node),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON5, NULL, NULL, NULL, leon_init, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON4, NULL, NULL, NULL, leon_init, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3, NULL, NULL, NULL, leon_init, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_LEON3FT, NULL, NULL, NULL, leon_init, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_APBUART, apbuart_apb_fixup, NULL,
+			NULL,NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_ETHMAC, NULL, NULL,     \
+			greth_fixup_prop, greth_init, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_I2CMST, NULL, NULL,     \
+			i2c_fixup_prop, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_SPICTRL, NULL, NULL,     \
+			spi_fixup_prop, NULL, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_GPIO, NULL, grgpio_fixup_core,     \
+			grgpio_fixup_prop, NULL, NULL),
  AMBAPP_DRV_EXTRA_FIXUP(VENDOR_GAISLER, GAISLER_CANAHB, NULL, NULL,     \
 			canoc_ahb_fixup_prop, NULL,                     \
-			canoc_ahb_fixup_dev_count, canoc_ahb_fixup_virtual),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_CANMUX, NULL, NULL, NULL, canmux_init),
- AMBAPP_DRV(VENDOR_GAISLER, GAISLER_USBDC, NULL, grusbdc_fixup_core, NULL, NULL),
+			canoc_ahb_fixup_dev_count, canoc_ahb_fixup_virtual, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_CANMUX, NULL, NULL,     \
+			NULL, canmux_init, NULL),
+ AMBAPP_DRV(VENDOR_GAISLER, GAISLER_USBDC, NULL, grusbdc_fixup_core,     \
+			NULL, NULL, NULL),
 };
 #define AMBAPP_DRV_NUM (sizeof(amba_drivers)/sizeof(struct amba_driver))
 
@@ -904,6 +1017,8 @@ int ambapp_fixup_prop(ambapp_core *core, struct prop_std *buf, int index)
 
 int ambapp_node_fixup(ambapp_core *core, struct node *node)
 {
+	struct amba_driver *drv;
+	int i;
 	ambapp_dev *dev = ambapp_get_first_dev(core);
 
 	if (dev->vendor == VENDOR_GAISLER) {
@@ -923,6 +1038,17 @@ int ambapp_node_fixup(ambapp_core *core, struct node *node)
 		    dev->device != GAISLER_GPTIMER)
 			return 1;
 #endif
+
+		for (i = 0, drv = amba_drivers; i < AMBAPP_DRV_NUM;
+		     i++, drv++) {
+			if ((drv->device == dev->device) &&
+			    (drv->fixup_node != NULL) &&
+			    (drv->vendor == dev->vendor) &&
+			    ((dev->ver >= drv->ver_min) &&
+			     (dev->ver <= drv->ver_max))) {
+				return drv->fixup_node(core, node);
+			}
+		}
 	}
 	return 0;
 }
@@ -2225,6 +2351,43 @@ void node_add_child(struct node *node, struct node *child)
 	*np = child;
 }
 
+/**
+ * node_has_child: Go through the children of the node and check if there is
+ * a child with the expected name.
+ *
+ * node       - The parent node
+ * child_name - The name of the child to be found
+ * name_len   - max length of the child_name
+ * out        - If provided it will point to the found child node
+
+ * Return values:
+ *  0 = No child found
+ *  1 = Found a child
+ */
+int node_has_child(struct node *node, const char *child_name, int name_len,
+		   struct node **out)
+{
+	struct node *n;
+	int proplen = _strnlen(child_name, name_len) + 1;
+	char tmp[proplen];
+
+	n = node->child;
+	while (n) {
+		/* Check if the length of the properties match, then go and
+		 * copy the property into tmp and then compare
+		 */
+		if (no_proplen((int)n, prop_str_name) == proplen &&
+			no_getprop((int)n, prop_str_name, tmp) == 1 &&
+			_strcmp(tmp, (char*)child_name) == 0) {
+			if (out)
+				*out = n;
+			return 1;
+		}
+		/* Go on and check the siblings */
+		n = n->sibling;
+	}
+	return 0;
+}
 
 #ifdef HASXML
 extern struct xml_match xml_matches[];
@@ -2429,17 +2592,13 @@ struct node *ambapp_node_create(ambapp_core *core)
 		size = ambapp_prop_creators[i](core, buf);
 		if (size == 0)
 			continue;
-		prop = (struct prop_std *)prom_malloc(size);
-		_memcpy((void *)prop, (void *)buf, size);
-		prop_add(n, (struct prop *)prop);
+		prop_add_dup(n, buf, size, &prop);
 	}
 	for (i = 0; ; i++) {
 		size = ambapp_fixup_prop(core, buf, i);
 		if (size == 0)
 			break;
-		prop = (struct prop_std *)prom_malloc(size);
-		_memcpy((void *)prop, (void *)buf, size);
-		prop_add(n, (struct prop *)prop);
+		prop_add_dup(n, buf, size, &prop);
 	}
 
 	if (xml_node_create_hook(core, n, &prop))
